"""

Protocol packet encoder code generator.

"""

import os
from collections import defaultdict

from . import protocol_declaration


def get_fmt_by_type(t):
    if isinstance(t, protocol_declaration.FloatArgument):
        fmt = "f"
    elif isinstance(t, protocol_declaration.DoubleArgument):
        fmt = "d"
    elif isinstance(t, protocol_declaration.BoolArgument):
        fmt = "b"
    elif isinstance(t, protocol_declaration.UInt8Argument):
        fmt = "B"
    elif isinstance(t, protocol_declaration.UInt16Argument):
        fmt = "H"
    elif isinstance(t, protocol_declaration.UInt32Argument):
        fmt = "L"
    elif isinstance(t, protocol_declaration.Int8Argument):
        fmt = "b"
    elif isinstance(t, protocol_declaration.Int16Argument):
        fmt = "h"
    elif isinstance(t, protocol_declaration.Int32Argument):
        fmt = "l"
    else:
        fmt = None
    return fmt


def get_farray_fmt(argument: protocol_declaration.FArrayArgument):
    data_fmt = get_fmt_by_type(argument.data_type)
    if not data_fmt:
        raise NotImplementedError("Unexpected farray data type '{}' for '{}'".format(
            argument.data_type.__class__.__name__, argument.name))
    return data_fmt


def get_varray_fmts(argument: protocol_declaration.VArrayArgument):
    length_fmt = get_fmt_by_type(argument.length_type)
    if not length_fmt:
        raise NotImplementedError("Unexpected varray length type '{}' for '{}'".format(
            argument.length_type.__class__.__name__, argument.name))
    data_fmt = get_fmt_by_type(argument.data_type)
    if not data_fmt:
        raise NotImplementedError("Unexpected varray data type '{}' for '{}'".format(
            argument.data_type.__class__.__name__, argument.name))
    return length_fmt, data_fmt


def get_string_fmt(argument: protocol_declaration.StringArgument):
    length_fmt = get_fmt_by_type(argument.length_type)
    if not length_fmt:
        raise NotImplementedError("Unexpected string length type '{}' for '{}'".format(
            argument.length_type.__class__.__name__, argument.name))
    return length_fmt


def get_enum_fmt(argument: protocol_declaration.EnumArgument):
    data_fmt = get_fmt_by_type(argument.data_type)
    if not data_fmt:
        raise NotImplementedError("Unexpected enum data type '{}' for '{}'".format(
            argument.data_type.__class__.__name__, argument.name))
    return data_fmt


def int_to_str(value: int, base: int = 10) -> str:
    if base == 8:
        res = "0o{:o}".format(value)
    elif base == 10:
        res = "{:d}".format(value)
    elif base == 16:
        res = "0x{:x}".format(value)
    else:
        raise ValueError("Unexpected base {}.".format(base))
    return res


class ProtocolGenerator(object):

    def __init__(self, f):
        self.f = f

    def generate_packet_slots(self, struct: protocol_declaration.Struct):
        for argument in struct.arguments:
            type_hint = argument.type_hint()
            if type_hint:
                type_hint = "  # {}".format(type_hint)
            else:
                type_hint = ""
            self.f.write('        "_{name}",{type_hint}\n'.format(name=argument.name, type_hint=type_hint))

    def generate_farray_validation(self, argument: protocol_declaration.FArrayArgument):
        if isinstance(argument.data_type, protocol_declaration.FloatArgument):
            element_validation = "lambda name, value_inner: validate_float(name, value_inner)"
        elif isinstance(argument.data_type, protocol_declaration.DoubleArgument):
            element_validation = "lambda name, value_inner: validate_float(name, value_inner)"
        elif isinstance(argument.data_type, protocol_declaration.BoolArgument):
            element_validation = "lambda name, value_inner: validate_bool(name, value_inner)"
        elif isinstance(argument.data_type, protocol_declaration.UInt8Argument):
            element_validation = "lambda name, value_inner: validate_integer(name, value_inner, 0, 255)"
        elif isinstance(argument.data_type, protocol_declaration.UInt16Argument):
            element_validation = "lambda name, value_inner: validate_integer(name, value_inner, 0, 65535)"
        elif isinstance(argument.data_type, protocol_declaration.UInt32Argument):
            element_validation = "lambda name, value_inner: validate_integer(name, value_inner, 0, 4294967295)"
        elif isinstance(argument.data_type, protocol_declaration.Int8Argument):
            element_validation = "lambda name, value_inner: validate_integer(name, value_inner, -128, 127)"
        elif isinstance(argument.data_type, protocol_declaration.Int16Argument):
            element_validation = "lambda name, value_inner: validate_integer(name, value_inner, -32768, 32767)"
        elif isinstance(argument.data_type, protocol_declaration.Int32Argument):
            element_validation = \
                "lambda name, value_inner: validate_integer(name, value_inner, -2147483648, 2147483647)"
        elif isinstance(argument.data_type, protocol_declaration.Struct):
            element_validation = "lambda name, value_inner: validate_object(name, value_inner, {})".format(
                argument.data_type.name)
        else:
            raise NotImplementedError("Unexpected farray data type '{}' for '{}'".format(
                argument.data_type.__class__.__name__, argument.name))

        self.f.write('validate_farray(\n            "{name}", value, {length}, {element_validation})\n'.format(
            name=argument.name, length=argument.length, element_validation=element_validation))

    def generate_varray_validation(self, argument: protocol_declaration.VArrayArgument):
        if isinstance(argument.length_type, protocol_declaration.UInt8Argument):
            maximum_length = 255
        elif isinstance(argument.length_type, protocol_declaration.UInt16Argument):
            maximum_length = 65535
        else:
            raise NotImplementedError("Unexpected varray length type '{}' for '{}'".format(
                argument.length_type.__class__.__name__, argument.name))

        if isinstance(argument.data_type, protocol_declaration.FloatArgument):
            element_validation = "lambda name, value_inner: validate_float(name, value_inner)"
        elif isinstance(argument.data_type, protocol_declaration.DoubleArgument):
            element_validation = "lambda name, value_inner: validate_float(name, value_inner)"
        elif isinstance(argument.data_type, protocol_declaration.BoolArgument):
            element_validation = "lambda name, value_inner: validate_bool(name, value_inner)"
        elif isinstance(argument.data_type, protocol_declaration.UInt8Argument):
            element_validation = "lambda name, value_inner: validate_integer(name, value_inner, 0, 255)"
        elif isinstance(argument.data_type, protocol_declaration.UInt16Argument):
            element_validation = "lambda name, value_inner: validate_integer(name, value_inner, 0, 65535)"
        elif isinstance(argument.data_type, protocol_declaration.UInt32Argument):
            element_validation = "lambda name, value_inner: validate_integer(name, value_inner, 0, 4294967295)"
        elif isinstance(argument.data_type, protocol_declaration.Int8Argument):
            element_validation = "lambda name, value_inner: validate_integer(name, value_inner, -128, 127)"
        elif isinstance(argument.data_type, protocol_declaration.Int16Argument):
            element_validation = "lambda name, value_inner: validate_integer(name, value_inner, -32768, 32767)"
        elif isinstance(argument.data_type, protocol_declaration.Int32Argument):
            element_validation = \
                "lambda name, value_inner: validate_integer(name, value_inner, -2147483648, 2147483647)"
        else:
            raise NotImplementedError("Unexpected varray data type '{}' for '{}'".format(
                argument.data_type.__class__.__name__, argument.name))

        self.f.write('validate_varray(\n            "{name}", value, {maximum_length}, {element_validation})\n'.format(
            name=argument.name, maximum_length=maximum_length, element_validation=element_validation))

    def generate_string_validation(self, argument: protocol_declaration.StringArgument):
        if isinstance(argument.length_type, protocol_declaration.UInt8Argument):
            maximum_length = 255
        elif isinstance(argument.length_type, protocol_declaration.UInt16Argument):
            maximum_length = 65535
        else:
            raise NotImplementedError("Unexpected string length type '{}' for '{}'".format(
                argument.length_type.__class__.__name__, argument.name))

        self.f.write('validate_string("{name}", value, {maximum_length})\n'.format(
            name=argument.name, maximum_length=maximum_length))

    def generate_enum_validation(self, argument: protocol_declaration.EnumArgument):
        if isinstance(argument.data_type, protocol_declaration.UInt8Argument):
            self.f.write('validate_integer("{name}", value.value, 0, 255)\n'.format(name=argument.name))
        elif isinstance(argument.data_type, protocol_declaration.UInt16Argument):
            self.f.write('validate_integer("{name}", value.value, 0, 65535)\n'.format(name=argument.name))
        elif isinstance(argument.data_type, protocol_declaration.UInt32Argument):
            self.f.write('validate_integer("{name}", value.value, 0, 4294967295)\n'.format(name=argument.name))
        elif isinstance(argument.data_type, protocol_declaration.Int8Argument):
            self.f.write('validate_integer("{name}", value.value, -128, 127)\n'.format(name=argument.name))
        elif isinstance(argument.data_type, protocol_declaration.Int16Argument):
            self.f.write('validate_integer("{name}", value.value, -32768, 32767)\n'.format(name=argument.name))
        elif isinstance(argument.data_type, protocol_declaration.Int32Argument):
            self.f.write(
                'validate_integer("{name}", value.value, -2147483648, 2147483647)\n'.format(name=argument.name))
        else:
            raise NotImplementedError("Unexpected enum data type '{}' for '{}'".format(
                argument.data_type.__class__.__name__, argument.name))

    def generate_argument_methods(self, struct: protocol_declaration.Struct):
        for argument in struct.arguments:
            if isinstance(argument, protocol_declaration.EnumArgument):
                self.f.write(r"""
    @property
    def {name}(self) -> {enum_type}:
        return self._{name}

    @{name}.setter
    def {name}(self, value: {enum_type}):
        self._{name} = value
        """.format(name=argument.name, enum_type=argument.enum_type.name))
                self.generate_enum_validation(argument)
            else:
                self.f.write(r"""
    @property
    def {name}(self):
        return self._{name}

    @{name}.setter
    def {name}(self, value):
        self._{name} = """.format(name=argument.name))
                if isinstance(argument, protocol_declaration.FloatArgument):
                    self.f.write('validate_float("{name}", value)\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.DoubleArgument):
                    self.f.write('validate_float("{name}", value)\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.BoolArgument):
                    self.f.write('validate_bool("{name}", value)\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.UInt8Argument):
                    self.f.write('validate_integer("{name}", value, 0, 255)\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.UInt16Argument):
                    self.f.write('validate_integer("{name}", value, 0, 65535)\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.UInt32Argument):
                    self.f.write('validate_integer("{name}", value, 0, 4294967295)\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.Int8Argument):
                    self.f.write('validate_integer("{name}", value, -128, 127)\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.Int16Argument):
                    self.f.write('validate_integer("{name}", value, -32768, 32767)\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.Int32Argument):
                    self.f.write(
                        'validate_integer("{name}", value, -2147483648, 2147483647)\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.FArrayArgument):
                    self.generate_farray_validation(argument)
                elif isinstance(argument, protocol_declaration.VArrayArgument):
                    self.generate_varray_validation(argument)
                elif isinstance(argument, protocol_declaration.StringArgument):
                    self.generate_string_validation(argument)
                else:
                    raise NotImplementedError("Unexpected argument type '{}' for '{}'".format(
                        argument.__class__.__name__, argument.name))

    def generate_len_method(self, struct: protocol_declaration.Struct):
        self.f.write("\n    def __len__(self):\n")
        if not struct.arguments:
            self.f.write("        return 0\n")
        else:
            self.f.write("        return \\\n")
            statements = []
            for argument in struct.arguments:
                if isinstance(argument, protocol_declaration.FloatArgument):
                    statements.append("get_size('f')")
                elif isinstance(argument, protocol_declaration.DoubleArgument):
                    statements.append("get_size('d')")
                elif isinstance(argument, protocol_declaration.BoolArgument):
                    statements.append("get_size('b')")
                elif isinstance(argument, protocol_declaration.UInt8Argument):
                    statements.append("get_size('B')")
                elif isinstance(argument, protocol_declaration.UInt16Argument):
                    statements.append("get_size('H')")
                elif isinstance(argument, protocol_declaration.UInt32Argument):
                    statements.append("get_size('L')")
                elif isinstance(argument, protocol_declaration.Int8Argument):
                    statements.append("get_size('b')")
                elif isinstance(argument, protocol_declaration.Int16Argument):
                    statements.append("get_size('h')")
                elif isinstance(argument, protocol_declaration.Int32Argument):
                    statements.append("get_size('l')")
                elif isinstance(argument, protocol_declaration.FArrayArgument):
                    if isinstance(argument.data_type, protocol_declaration.Struct):
                        statements.append("get_object_farray_size(self._{name}, {length})".format(
                            name=argument.name, length=argument.length))
                    else:
                        data_fmt = get_farray_fmt(argument)
                        statements.append("get_farray_size('{data_fmt}', {length})".format(
                            data_fmt=data_fmt, length=argument.length))
                elif isinstance(argument, protocol_declaration.VArrayArgument):
                    length_fmt, data_fmt = get_varray_fmts(argument)
                    statements.append("get_varray_size(self._{name}, '{length_fmt}', '{data_fmt}')".format(
                        name=argument.name, length_fmt=length_fmt, data_fmt=data_fmt))
                elif isinstance(argument, protocol_declaration.StringArgument):
                    length_fmt = get_string_fmt(argument)
                    statements.append("get_string_size(self._{name}, '{length_fmt}')".format(
                        name=argument.name, length_fmt=length_fmt))
                elif isinstance(argument, protocol_declaration.EnumArgument):
                    data_fmt = get_enum_fmt(argument)
                    statements.append("get_size('{}')".format(data_fmt))
                else:
                    raise NotImplementedError("Unexpected argument type '{}' for '{}'".format(
                        argument.__class__.__name__, argument.name))
            self.f.write("            ")
            self.f.write(" + \\\n            ".join(statements))
            self.f.write("\n")

    def generate_repr_method(self, struct: protocol_declaration.Struct):
        self.f.write("\n    def __repr__(self):\n")
        if struct.arguments:
            argument_strs = []
            arguments = ["type=type(self).__name__"]
            if struct.arguments:
                for argument in struct.arguments:
                    argument_strs.append('"{name}={{{name}}}'.format(name=argument.name))
                    arguments.append("{name}=self._{name}".format(name=argument.name))
            self.f.write('        return "{type}(" \\\n               ')
            self.f.write('{argument_strs})".format(\n                {arguments})\n'.format(
                argument_strs=', " \\\n               '.join(argument_strs),
                arguments=",\n                ".join(arguments)))
        else:
            self.f.write('        return "{type}()".format(type=type(self).__name__)\n')

    def generate_argument_defaults(self, struct: protocol_declaration.Struct):
        for argument in struct.arguments:
            if isinstance(argument.default, str):
                self.f.write(
                    ",\n                 {name}='{default}'".format(name=argument.name, default=argument.default))
            else:
                self.f.write(
                    ",\n                 {name}={default}".format(name=argument.name, default=argument.default))

    def generate_arugment_assignments(self, struct: protocol_declaration.Struct):
        if struct.arguments:
            for argument in struct.arguments:
                if isinstance(argument, protocol_declaration.EnumArgument):
                    self.f.write("        self.{name} = {enum_type}({name})\n".format(
                        name=argument.name, enum_type=argument.enum_type.name))
                else:
                    self.f.write("        self.{name} = {name}\n".format(name=argument.name))
        else:
            self.f.write("        pass\n")

    def generate_packet_arugment_assignments(self, packet: protocol_declaration.Packet):
        packet_id = "0x{:02x}".format(packet.id) if packet.id is not None else None
        self.f.write("        super().__init__({type}, packet_id={id})\n".format(type=packet.type, id=packet_id))
        self.generate_arugment_assignments(packet)

    def generate_packet_encoding(self, struct: protocol_declaration.Struct):
        self.f.write(r"""
    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
""")
        if struct.arguments:
            for argument in struct.arguments:
                if isinstance(argument, protocol_declaration.FloatArgument):
                    self.f.write('        writer.write(self._{name}, "f")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.DoubleArgument):
                    self.f.write('        writer.write(self._{name}, "d")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.BoolArgument):
                    self.f.write('        writer.write(int(self._{name}), "b")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.UInt8Argument):
                    self.f.write('        writer.write(self._{name}, "B")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.UInt16Argument):
                    self.f.write('        writer.write(self._{name}, "H")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.UInt32Argument):
                    self.f.write('        writer.write(self._{name}, "L")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.Int8Argument):
                    self.f.write('        writer.write(self._{name}, "b")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.Int16Argument):
                    self.f.write('        writer.write(self._{name}, "h")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.Int32Argument):
                    self.f.write('        writer.write(self._{name}, "l")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.FArrayArgument):
                    if isinstance(argument.data_type, protocol_declaration.Struct):
                        self.f.write('        writer.write_object_farray(self._{name}, {length})\n'.format(
                            name=argument.name, length=argument.length))
                    else:
                        data_fmt = get_farray_fmt(argument)
                        self.f.write('        writer.write_farray(self._{name}, "{data_fmt}", {length})\n'.format(
                            name=argument.name, length=argument.length, data_fmt=data_fmt))
                elif isinstance(argument, protocol_declaration.VArrayArgument):
                    length_fmt, data_fmt = get_varray_fmts(argument)
                    self.f.write('        writer.write_varray(self._{name}, "{data_fmt}", "{length_fmt}")\n'.format(
                        name=argument.name, length_fmt=length_fmt, data_fmt=data_fmt))
                elif isinstance(argument, protocol_declaration.StringArgument):
                    length_fmt = get_string_fmt(argument)
                    self.f.write('        writer.write_string(self._{name}, "{length_fmt}")\n'.format(
                        name=argument.name, length_fmt=length_fmt))
                elif isinstance(argument, protocol_declaration.EnumArgument):
                    data_fmt = get_enum_fmt(argument)
                    self.f.write('        writer.write(self._{name}.value, "{data_fmt}")\n'.format(
                        name=argument.name, data_fmt=data_fmt))
                else:
                    raise NotImplementedError("Unexpected argument type '{}' for '{}'".format(
                        argument.__class__.__name__, argument.name))
        else:
            self.f.write("        pass\n")

    def generate_packet_decoding(self, struct: protocol_declaration.Struct):
        self.f.write(r"""
    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
""")
        if struct.arguments:
            for argument in struct.arguments:
                if isinstance(argument, protocol_declaration.FloatArgument):
                    self.f.write('        {name} = reader.read("f")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.DoubleArgument):
                    self.f.write('        {name} = reader.read("d")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.BoolArgument):
                    self.f.write('        {name} = bool(reader.read("b"))\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.UInt8Argument):
                    self.f.write('        {name} = reader.read("B")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.UInt16Argument):
                    self.f.write('        {name} = reader.read("H")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.UInt32Argument):
                    self.f.write('        {name} = reader.read("L")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.Int8Argument):
                    self.f.write('        {name} = reader.read("b")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.Int16Argument):
                    self.f.write('        {name} = reader.read("h")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.Int32Argument):
                    self.f.write('        {name} = reader.read("l")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.FArrayArgument):
                    if isinstance(argument.data_type, protocol_declaration.Struct):
                        self.f.write(
                            '        {name} = reader.read_object_farray({data_type}.from_reader, {length})\n'.format(
                                name=argument.name, data_type=argument.data_type.name, length=argument.length))
                    else:
                        data_fmt = get_farray_fmt(argument)
                        self.f.write('        {name} = reader.read_farray("{data_fmt}", {length})\n'.format(
                            name=argument.name, length=argument.length, data_fmt=data_fmt))
                elif isinstance(argument, protocol_declaration.VArrayArgument):
                    length_fmt, data_fmt = get_varray_fmts(argument)
                    self.f.write('        {name} = reader.read_varray("{data_fmt}", "{length_fmt}")\n'.format(
                        name=argument.name, length_fmt=length_fmt, data_fmt=data_fmt))
                elif isinstance(argument, protocol_declaration.StringArgument):
                    length_fmt = get_string_fmt(argument)
                    self.f.write('        {name} = reader.read_string("{length_fmt}")\n'.format(
                        name=argument.name, length_fmt=length_fmt))
                elif isinstance(argument, protocol_declaration.EnumArgument):
                    data_fmt = get_enum_fmt(argument)
                    self.f.write('        {name} = reader.read("{data_fmt}")\n'.format(
                        name=argument.name, data_fmt=data_fmt))
                else:
                    raise NotImplementedError("Unexpected argument type '{}' for '{}'".format(
                        argument.__class__.__name__, argument.name))
        else:
            self.f.write("        del reader\n")
        self.f.write("        return cls(\n")
        arguments = []
        for argument in struct.arguments:
            arguments.append("{name}={name}".format(name=argument.name))
        self.f.write("            ")
        self.f.write(",\n            ".join(arguments))
        self.f.write(")\n")

    def generate_enum(self, enum_type: protocol_declaration.Enum):

        self.f.write(r"""

class {name}(enum.Enum):
""".format(name=enum_type.name))
        for member in enum_type.members:
            self.f.write('    {name} = {value}\n'.format(
                name=member.name, value=int_to_str(member.value, enum_type.base)))

    def generate_struct(self, struct: protocol_declaration.Struct):
        self.f.write(r"""

class {name}(Struct):
""".format(name=struct.name))
        self.f.write("\n    __slots__ = (\n")
        self.generate_packet_slots(struct)
        self.f.write("    )\n\n    def __init__(self")
        self.generate_argument_defaults(struct)
        self.f.write("):\n")
        self.generate_arugment_assignments(struct)
        self.generate_argument_methods(struct)
        self.generate_len_method(struct)
        self.generate_repr_method(struct)
        self.generate_packet_encoding(struct)
        self.generate_packet_decoding(struct)

    def generate_packet(self, packet):
        self.f.write(r"""
    
class {name}(Packet):
""".format(name=packet.name))
        self.f.write("\n    __slots__ = (\n")
        self.generate_packet_slots(packet)
        self.f.write("    )\n\n    def __init__(self")
        self.generate_argument_defaults(packet)
        self.f.write("):\n")
        self.generate_packet_arugment_assignments(packet)
        self.generate_argument_methods(packet)
        self.generate_len_method(packet)
        self.generate_repr_method(packet)
        self.generate_packet_encoding(packet)
        self.generate_packet_decoding(packet)

    def generate_id_map(self):
        packet_map = {}
        for pkt in protocol_declaration.PROTOCOL.packets:
            if pkt.id:
                packet_map[pkt.id] = pkt.name

        self.f.write('\n\nPACKETS_BY_ID = {\n')
        for k, v in sorted(packet_map.items()):
            self.f.write('    0x{id:02x}: {name},  # {id}\n'.format(id=k, name=v))
        self.f.write('}\n')

    def generate_group_map(self):
        packet_map = defaultdict(set)
        for pkt in protocol_declaration.PROTOCOL.packets:
            if pkt.group and pkt.id:
                packet_map[pkt.group].add(pkt)

        self.f.write('\n\nPACKETS_BY_GROUP = {\n')
        for group, pkt_set in sorted(packet_map.items()):
            self.f.write('    "{group}": {{\n'.format(group=group))
            for pkt in sorted(pkt_set, key=lambda pkt2: pkt2.id):
                self.f.write('        0x{id:02x},  # {name}\n'.format(id=pkt.id, name=pkt.name))
            self.f.write('    },\n')
        self.f.write('}\n')

    def generate(self):
        header = r'''"""

Protocol packet encoder classes.

Generated from {declaration} by {generator}

Do not modify.

"""

import enum

from .protocol_declaration import PacketType
from .protocol_base import Struct, Packet
from .protocol_utils import \
    validate_float, validate_bool, validate_integer, validate_object, \
    validate_farray, validate_varray, validate_string, \
    get_size, get_farray_size, get_varray_size, get_string_size, get_object_farray_size, \
    BinaryReader, BinaryWriter
'''.format(declaration=os.path.basename(protocol_declaration.__file__), generator=os.path.basename(__file__))
        self.f.write(header)

        for enum in protocol_declaration.PROTOCOL.enums:
            self.generate_enum(enum)

        for struct in protocol_declaration.PROTOCOL.structs:
            self.generate_struct(struct)

        for packet in protocol_declaration.PROTOCOL.packets:
            self.generate_packet(packet)

        self.generate_id_map()
        self.generate_group_map()




import math

from . import util
from . import protocol_encoder


MIN_HEAD_ANGLE = util.Angle(degrees=-25)
MAX_HEAD_ANGLE = util.Angle(degrees=44.5)

MIN_LIFT_HEIGHT = util.Distance(mm=32.0)
MAX_LIFT_HEIGHT = util.Distance(mm=92.0)

LIFT_ARM_LENGTH = util.Distance(mm=66.0)
LIFT_PIVOT_HEIGHT = util.Distance(mm=45.0)

MIN_LIFT_ANGLE = util.Angle(radians=math.asin((MIN_LIFT_HEIGHT.mm - LIFT_PIVOT_HEIGHT.mm) / LIFT_ARM_LENGTH.mm))
MAX_LIFT_ANGLE = util.Angle(radians=math.asin((MAX_LIFT_HEIGHT.mm - LIFT_PIVOT_HEIGHT.mm) / LIFT_ARM_LENGTH.mm))

MAX_WHEEL_SPEED = util.Speed(mmps=200.0)

TRACK_WIDTH = util.Distance(mm=45.0)


class RobotStatusFlag(object):
    IS_MOVING = 0x1
    IS_CARRYING_BLOCK = 0x2
    IS_PICKING_OR_PLACING = 0x4
    IS_PICKED_UP = 0x8
    IS_BODY_ACC_MODE = 0x10
    IS_FALLING = 0x20
    IS_ANIMATING = 0x40
    IS_PATHING = 0x80
    LIFT_IN_POS = 0x100
    HEAD_IN_POS = 0x200
    IS_ANIM_BUFFER_FULL = 0x400
    IS_ANIMATING_IDLE = 0x800
    IS_ON_CHARGER = 0x1000
    IS_CHARGING = 0x2000
    CLIFF_DETECTED = 0x4000
    ARE_WHEELS_MOVING = 0x8000
    IS_CHARGER_OOS = 0x10000


RobotStatusFlagNames = {
    RobotStatusFlag.IS_MOVING: "IS_MOVING",
    RobotStatusFlag.IS_CARRYING_BLOCK: "IS_CARRYING_BLOCK",
    RobotStatusFlag.IS_PICKING_OR_PLACING: "IS_PICKING_OR_PLACING",
    RobotStatusFlag.IS_PICKED_UP: "IS_PICKED_UP",
    RobotStatusFlag.IS_BODY_ACC_MODE: "IS_BODY_ACC_MODE",
    RobotStatusFlag.IS_FALLING: "IS_FALLING",
    RobotStatusFlag.IS_ANIMATING: "IS_ANIMATING",
    RobotStatusFlag.IS_PATHING: "IS_PATHING",
    RobotStatusFlag.LIFT_IN_POS: "LIFT_IN_POS",
    RobotStatusFlag.HEAD_IN_POS: "HEAD_IN_POS",
    RobotStatusFlag.IS_ANIM_BUFFER_FULL: "IS_ANIM_BUFFER_FULL",
    RobotStatusFlag.IS_ANIMATING_IDLE: "IS_ANIMATING_IDLE",
    RobotStatusFlag.IS_ON_CHARGER: "IS_ON_CHARGER",
    RobotStatusFlag.IS_CHARGING: "IS_CHARGING",
    RobotStatusFlag.CLIFF_DETECTED: "CLIFF_DETECTED",
    RobotStatusFlag.ARE_WHEELS_MOVING: "ARE_WHEELS_MOVING",
    RobotStatusFlag.IS_CHARGER_OOS: "IS_CHARGER_OOS",
}


BODY_COLOR_NAMES = {
    protocol_encoder.BodyColor.WHITE_v10: "Original",
    protocol_encoder.BodyColor.RESERVED: "Reserved",
    protocol_encoder.BodyColor.WHITE_v15: "White",
    protocol_encoder.BodyColor.CE_LM_v15: "CE_LM",
    protocol_encoder.BodyColor.LE_BL_v16: "LE_BL",
}


class LiftPosition(object):
    """
    Represents the position of Cozmo's lift.

    The class allows the position to be referred to as either absolute height
    above the ground, as a ratio from 0.0 to 1.0, or as the angle of the lift
    arm relative to the ground.

    Args:
        height (:class:`cozmo.util.Distance`): The height of the lift above the ground.
        ratio (float): The ratio from 0.0 to 1.0 that the lift is raised from the ground.
        angle (:class:`cozmo.util.Angle`): The angle of the lift arm relative to the ground.
    """

    __slots__ = ('_height', )

    def __init__(self, height=None, ratio=None, angle=None):
        def _count_arg(arg):
            # return 1 if argument is set (not None), 0 otherwise
            return 0 if (arg is None) else 1
        num_provided_args = _count_arg(height) + _count_arg(ratio) + _count_arg(angle)
        if num_provided_args != 1:
            raise ValueError("Expected one, and only one, of the distance, ratio or angle keyword arguments")

        if height is not None:
            if not isinstance(height, util.Distance):
                raise TypeError("Unsupported type for distance - expected util.Distance")
            self._height = height
        elif ratio is not None:
            height_mm = MIN_LIFT_HEIGHT.mm + (ratio * (MAX_LIFT_HEIGHT.mm - MIN_LIFT_HEIGHT.mm))
            self._height = util.Distance(mm=height_mm)
        elif angle is not None:
            if not isinstance(angle, util.Angle):
                raise TypeError("Unsupported type for angle - expected util.Angle")
            height_mm = (math.sin(angle.radians) * LIFT_ARM_LENGTH.mm) + LIFT_PIVOT_HEIGHT.mm
            self._height = util.Distance(mm=height_mm)

    def __repr__(self):
        return "<%s height=%s ratio=%s angle=%s>" % (self.__class__.__name__, self._height, self.ratio, self.angle)

    @property
    def height(self) -> util.Distance:
        """ Height above the ground. """
        return self._height

    @property
    def ratio(self) -> float:
        """ The ratio from 0 to 1 that the lift is raised, 0 at the bottom, 1 at the top. """
        ratio = ((self._height.mm - MIN_LIFT_HEIGHT.mm) / (MAX_LIFT_HEIGHT.mm - MIN_LIFT_HEIGHT.mm))
        return ratio

    @property
    def angle(self) -> util.Angle:
        """ The angle of the lift arm relative to the ground. """
        sin_angle = (self._height.mm - LIFT_PIVOT_HEIGHT.mm) / LIFT_ARM_LENGTH.mm
        angle = math.asin(sin_angle)
        return util.Angle(radians=angle)




__all__ = [
    "PyCozmoException",
]


class PyCozmoException(Exception):
    """ Base class for all PyCozmo exceptions. """


class PyCozmoConnectionError(PyCozmoException):
    """ Base class for all PyCozmo connection exceptions. """


class ConnectionTimeout(PyCozmoConnectionError):
    """ Connection timeout. """


class UnsupportedFirmwareVersion(PyCozmoConnectionError):
    """ Unsupported Cozmo firmware version. """


class Timeout(PyCozmoException):
    """ Timeout. """



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: CozmoAnim

import flatbuffers

class ProceduralFace(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsProceduralFace(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ProceduralFace()
        x.Init(buf, n + offset)
        return x

    # ProceduralFace
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ProceduralFace
    def TriggerTimeMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # ProceduralFace
    def FaceAngle(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # ProceduralFace
    def FaceCenterX(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # ProceduralFace
    def FaceCenterY(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # ProceduralFace
    def FaceScaleX(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 1.0

    # ProceduralFace
    def FaceScaleY(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 1.0

    # ProceduralFace
    def LeftEye(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # ProceduralFace
    def LeftEyeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # ProceduralFace
    def LeftEyeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ProceduralFace
    def RightEye(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # ProceduralFace
    def RightEyeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # ProceduralFace
    def RightEyeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def ProceduralFaceStart(builder): builder.StartObject(8)
def ProceduralFaceAddTriggerTimeMs(builder, triggerTimeMs): builder.PrependUint32Slot(0, triggerTimeMs, 0)
def ProceduralFaceAddFaceAngle(builder, faceAngle): builder.PrependFloat32Slot(1, faceAngle, 0.0)
def ProceduralFaceAddFaceCenterX(builder, faceCenterX): builder.PrependFloat32Slot(2, faceCenterX, 0.0)
def ProceduralFaceAddFaceCenterY(builder, faceCenterY): builder.PrependFloat32Slot(3, faceCenterY, 0.0)
def ProceduralFaceAddFaceScaleX(builder, faceScaleX): builder.PrependFloat32Slot(4, faceScaleX, 1.0)
def ProceduralFaceAddFaceScaleY(builder, faceScaleY): builder.PrependFloat32Slot(5, faceScaleY, 1.0)
def ProceduralFaceAddLeftEye(builder, leftEye): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(leftEye), 0)
def ProceduralFaceStartLeftEyeVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ProceduralFaceAddRightEye(builder, rightEye): builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(rightEye), 0)
def ProceduralFaceStartRightEyeVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ProceduralFaceEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: CozmoAnim

import flatbuffers

class BackpackLights(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsBackpackLights(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = BackpackLights()
        x.Init(buf, n + offset)
        return x

    # BackpackLights
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # BackpackLights
    def TriggerTimeMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # BackpackLights
    def DurationTimeMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # BackpackLights
    def Left(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # BackpackLights
    def LeftAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # BackpackLights
    def LeftLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # BackpackLights
    def Right(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # BackpackLights
    def RightAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # BackpackLights
    def RightLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # BackpackLights
    def Front(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # BackpackLights
    def FrontAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # BackpackLights
    def FrontLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # BackpackLights
    def Middle(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # BackpackLights
    def MiddleAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # BackpackLights
    def MiddleLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # BackpackLights
    def Back(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # BackpackLights
    def BackAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # BackpackLights
    def BackLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def BackpackLightsStart(builder): builder.StartObject(7)
def BackpackLightsAddTriggerTimeMs(builder, triggerTimeMs): builder.PrependUint32Slot(0, triggerTimeMs, 0)
def BackpackLightsAddDurationTimeMs(builder, durationTimeMs): builder.PrependUint32Slot(1, durationTimeMs, 0)
def BackpackLightsAddLeft(builder, Left): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(Left), 0)
def BackpackLightsStartLeftVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def BackpackLightsAddRight(builder, Right): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(Right), 0)
def BackpackLightsStartRightVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def BackpackLightsAddFront(builder, Front): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(Front), 0)
def BackpackLightsStartFrontVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def BackpackLightsAddMiddle(builder, Middle): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(Middle), 0)
def BackpackLightsStartMiddleVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def BackpackLightsAddBack(builder, Back): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(Back), 0)
def BackpackLightsStartBackVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def BackpackLightsEnd(builder): return builder.EndObject()




from typing import Optional, List

from PIL import Image, ImageDraw


WIDTH = 128
HEIGHT = 64

EYE_WIDTH = 28
EYE_HEIGHT = 40
HALF_EYE_WIDTH = EYE_WIDTH // 2
HALF_EYE_HEIGHT = EYE_HEIGHT // 2

RESAMPLE = Image.NEAREST


class ProceduralLid(object):

    BLACK = Image.new("1", (WIDTH * 2, HEIGHT * 2), color=0)

    def __init__(self,
                 offset: int = 0,
                 angle_offset: float = 0.0,
                 y: float = 0.0,
                 angle: float = 0.0,
                 bend: float = 0.0):
        self.offset = int(offset)
        self.angle_offset = float(angle_offset)
        self.y = float(y)
        self.angle = float(angle)
        self.bend = float(bend)

    def render(self, im: Image) -> None:
        # Lid image
        lid = Image.new("1", (WIDTH * 2, HEIGHT * 2), color=0)

        draw = ImageDraw.Draw(lid)

        # Draw lid
        lid_height = int(EYE_HEIGHT * self.y)
        x1 = WIDTH - EYE_WIDTH
        y1 = HEIGHT - 1 - HALF_EYE_HEIGHT
        x2 = WIDTH + EYE_WIDTH
        y2 = HEIGHT - 1 + lid_height
        draw.rectangle(((x1, y1), (x2, y2)), fill=1)

        bend_height = int(EYE_HEIGHT * (1.0 - self.y) * self.bend)
        x3 = WIDTH - HALF_EYE_WIDTH
        y3 = HEIGHT - 1 + lid_height - bend_height
        x4 = WIDTH + HALF_EYE_WIDTH
        y4 = HEIGHT - 1 + lid_height + bend_height
        draw.chord(((x3, y3), (x4, y4)), 0, 180, fill=1)

        # Rotate
        lid = lid.rotate(self.angle + self.angle_offset, resample=RESAMPLE, expand=0)

        # Translate and compose
        location = ((im.size[0] - lid.size[0]) // 2,
                    (im.size[1] - lid.size[1]) // 2 + self.offset)
        im.paste(self.BLACK, location, lid)


class ProceduralEye(object):

    CORNER_RADIUS = 16
    X_FACTOR = 0.55
    Y_FACTOR = 0.25

    def __init__(self,
                 offset: int = 0,
                 center_x: int = 0,
                 center_y: int = 0,
                 scale_x: float = 1.0,
                 scale_y: float = 1.0,
                 angle: float = 0.0,
                 lower_inner_radius_x: float = 0.5,
                 lower_inner_radius_y: float = 0.5,
                 lower_outer_radius_x: float = 0.5,
                 lower_outer_radius_y: float = 0.5,
                 upper_inner_radius_x: float = 0.5,
                 upper_inner_radius_y: float = 0.5,
                 upper_outer_radius_x: float = 0.5,
                 upper_outer_radius_y: float = 0.5,
                 upper_lid_y: float = 0.0,
                 upper_lid_angle: float = 0.0,
                 upper_lid_bend: float = 0.0,
                 lower_lid_y: float = 0.0,
                 lower_lid_angle: float = 0.0,
                 lower_lid_bend: float = 0.0):
        self.offset = int(offset)
        self.center_x = int(center_x)
        self.center_y = int(center_y)
        self.scale_x = float(scale_x)
        self.scale_y = float(scale_y)
        self.angle = float(angle)
        self.lower_inner_radius_x = float(lower_inner_radius_x)
        self.lower_inner_radius_y = float(lower_inner_radius_y)
        self.lower_outer_radius_x = float(lower_outer_radius_x)
        self.lower_outer_radius_y = float(lower_outer_radius_y)
        self.upper_inner_radius_x = float(upper_inner_radius_x)
        self.upper_inner_radius_y = float(upper_inner_radius_y)
        self.upper_outer_radius_x = float(upper_outer_radius_x)
        self.upper_outer_radius_y = float(upper_outer_radius_y)
        self.lids = (
            ProceduralLid(-HALF_EYE_HEIGHT, 0.0, upper_lid_y, upper_lid_angle, upper_lid_bend),
            ProceduralLid(HALF_EYE_HEIGHT + 1, 180.0, lower_lid_y, lower_lid_angle, lower_lid_bend)
        )

    def _render_inner_rect(self, draw, y1, x2, y2) -> None:
        x3 = x2 - int(self.CORNER_RADIUS * max(self.upper_inner_radius_x, self.lower_inner_radius_x))
        y3 = y1 + int(self.CORNER_RADIUS * self.upper_inner_radius_y)
        x4 = x2
        y4 = y2 - int(self.CORNER_RADIUS * self.lower_inner_radius_y)
        draw.rectangle(((x3, y3), (x4, y4)), fill=1)

    def _render_upper_rect(self, draw, x1, y1, x2) -> None:
        x3 = x1 + int(self.CORNER_RADIUS * self.upper_outer_radius_x)
        y3 = y1
        x4 = x2 - int(self.CORNER_RADIUS * self.upper_inner_radius_x)
        y4 = y1 + int(self.CORNER_RADIUS * max(self.upper_outer_radius_y, self.upper_inner_radius_y))
        draw.rectangle(((x3, y3), (x4, y4)), fill=1)

    def _render_outer_rect(self, draw, x1, y1, y2) -> None:
        x3 = x1
        y3 = y1 + int(self.CORNER_RADIUS * self.upper_outer_radius_y)
        x4 = x1 + int(self.CORNER_RADIUS * max(self.upper_outer_radius_x, self.lower_outer_radius_x))
        y4 = y2 - int(self.CORNER_RADIUS * self.lower_outer_radius_y)
        draw.rectangle(((x3, y3), (x4, y4)), fill=1)

    def _render_lower_rect(self, draw, x1, x2, y2) -> None:
        x3 = x1 + int(self.CORNER_RADIUS * self.lower_outer_radius_x)
        y3 = y2 - int(self.CORNER_RADIUS * max(self.lower_outer_radius_y, self.lower_inner_radius_y))
        x4 = x2 - int(self.CORNER_RADIUS * self.lower_inner_radius_x)
        y4 = y2
        draw.rectangle(((x3, y3), (x4, y4)), fill=1)

    def _render_center_rect(self, draw, x1, y1, x2, y2) -> None:
        x3 = x1 + int(self.CORNER_RADIUS * max(self.upper_outer_radius_x, self.lower_outer_radius_x))
        y3 = y1 + int(self.CORNER_RADIUS * max(self.upper_outer_radius_y, self.upper_inner_radius_y))
        x4 = x2 - int(self.CORNER_RADIUS * max(self.upper_inner_radius_y, self.lower_inner_radius_y))
        y4 = y2 - int(self.CORNER_RADIUS * max(self.lower_outer_radius_y, self.lower_inner_radius_y))
        draw.rectangle(((x3, y3), (x4, y4)), fill=1)

    def _render_lower_inner_pie(self, draw, x2, y2) -> None:
        x3 = x2 - 2 * int(self.CORNER_RADIUS * self.lower_inner_radius_x)
        y3 = y2 - 2 * int(self.CORNER_RADIUS * self.lower_inner_radius_y)
        x4 = x2
        y4 = y2
        draw.pieslice(((x3, y3), (x4, y4)), 0, 90, fill=1)

    def _render_upper_inner_pie(self, draw, y1, x2) -> None:
        x3 = x2 - 2 * int(self.CORNER_RADIUS * self.upper_inner_radius_x)
        y3 = y1
        x4 = x2
        y4 = y1 + 2 * int(self.CORNER_RADIUS * self.upper_inner_radius_y)
        draw.pieslice(((x3, y3), (x4, y4)), 270, 360, fill=1)

    def _render_upper_outer_pie(self, draw, x1, y1) -> None:
        x3 = x1
        y3 = y1
        x4 = x1 + 2 * int(self.CORNER_RADIUS * self.upper_outer_radius_x)
        y4 = y1 + 2 * int(self.CORNER_RADIUS * self.upper_outer_radius_y)
        draw.pieslice(((x3, y3), (x4, y4)), 180, 270, fill=1)

    def _render_lower_outer_pie(self, draw, x1, y2) -> None:
        x3 = x1
        y3 = y2 - 2 * int(self.CORNER_RADIUS * self.lower_outer_radius_y)
        x4 = x1 + 2 * int(self.CORNER_RADIUS * self.lower_outer_radius_x)
        y4 = y2
        draw.pieslice(((x3, y3), (x4, y4)), 90, 180, fill=1)

    def render(self, im: Image) -> None:
        # Eye image
        eye = Image.new("1", (WIDTH, HEIGHT), color=0)

        draw = ImageDraw.Draw(eye)
        x1 = eye.size[0] // 2 - HALF_EYE_WIDTH
        y1 = eye.size[1] // 2 - HALF_EYE_HEIGHT
        x2 = eye.size[0] // 2 + HALF_EYE_WIDTH
        y2 = eye.size[1] // 2 + HALF_EYE_HEIGHT
        self._render_inner_rect(draw, y1, x2, y2)
        self._render_upper_rect(draw, x1, y1, x2)
        self._render_outer_rect(draw, x1, y1, y2)
        self._render_lower_rect(draw, x1, x2, y2)
        self._render_center_rect(draw, x1, y1, x2, y2)
        self._render_lower_inner_pie(draw, x2, y2)
        self._render_upper_inner_pie(draw, y1, x2)
        self._render_upper_outer_pie(draw, x1, y1)
        self._render_lower_outer_pie(draw, x1, y2)

        # Draw lids
        self.lids[0].render(eye)
        self.lids[1].render(eye)

        # Rotate
        eye = eye.rotate(self.angle, resample=RESAMPLE, expand=1)

        # Scale
        scale = (int(float(eye.size[0]) * self.scale_x),
                 int(float(eye.size[1]) * self.scale_y))
        try:
            eye = eye.resize(scale, resample=RESAMPLE)
        except ValueError:
            # Scale factors can be extremely small and Pillow cannot handle resize() with both scale factors of 0.
            eye = None

        # Translate and compose
        if eye:
            location = ((im.size[0] - eye.size[0]) // 2 + int(self.center_x * self.X_FACTOR) + self.offset,
                        (im.size[1] - eye.size[1]) // 2 + int(self.center_y * self.Y_FACTOR))
            im.paste(eye, location, eye)


class ProceduralFace(object):

    X_FACTOR = 0.55
    Y_FACTOR = 0.25
    LEFT_EYE_OFFSET = -23
    RIGHT_EYE_OFFSET = 21

    def __init__(self,
                 center_x: int = 0,
                 center_y: int = 0,
                 scale_x: float = 1.0,
                 scale_y: float = 1.0,
                 angle: float = 0.0,
                 left_eye: Optional[List] = None,
                 right_eye: Optional[List] = None):
        self.center_x = int(center_x)
        self.center_y = int(center_y)
        self.scale_x = float(scale_x)
        self.scale_y = float(scale_y)
        self.angle = float(angle)
        if left_eye:
            self.left_eye = ProceduralEye(
                self.LEFT_EYE_OFFSET,
                left_eye[0], left_eye[1],
                left_eye[2], left_eye[3],
                left_eye[4],
                left_eye[5], left_eye[6],
                left_eye[7], left_eye[8],
                left_eye[9], left_eye[10],
                left_eye[11], left_eye[12],
                left_eye[13], left_eye[14], left_eye[15],
                left_eye[16], left_eye[17], left_eye[18],
            )
        else:
            self.left_eye = ProceduralEye(self.LEFT_EYE_OFFSET)
        if right_eye:
            self.right_eye = ProceduralEye(
                self.RIGHT_EYE_OFFSET,
                right_eye[0], right_eye[1],
                right_eye[2], right_eye[3],
                right_eye[4],
                right_eye[5], right_eye[6],
                right_eye[7], right_eye[8],
                right_eye[9], right_eye[10],
                right_eye[11], right_eye[12],
                right_eye[13], right_eye[14], right_eye[15],
                right_eye[16], right_eye[17], right_eye[18],
            )
        else:
            self.right_eye = ProceduralEye(self.RIGHT_EYE_OFFSET)

    def render(self) -> Image:
        # Background image
        im = Image.new("1", (WIDTH, HEIGHT), color=0)

        # Face image
        face = Image.new("1", (WIDTH, HEIGHT), color=0)

        # Draw eyes
        self.left_eye.render(face)
        self.right_eye.render(face)

        # Rotate
        face = face.rotate(self.angle, resample=RESAMPLE, expand=1)

        # Scale
        scale = (int(float(face.size[0]) * self.scale_x),
                 int(float(face.size[1] * self.scale_y)))
        try:
            face = face.resize(scale, resample=RESAMPLE)
        except ValueError:
            # Scale factors can be extremely small and Pillow cannot handle resize() with both scale factors of 0.
            face = None

        # Translate and compose
        if face:
            location = ((im.size[0] - face.size[0]) // 2 + int(self.center_x * self.X_FACTOR),
                        (im.size[1] - face.size[1]) // 2 + int(self.center_y * self.Y_FACTOR))
            im.paste(face, location)

        return im



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: CozmoAnim

import flatbuffers

class FaceAnimation(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsFaceAnimation(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = FaceAnimation()
        x.Init(buf, n + offset)
        return x

    # FaceAnimation
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # FaceAnimation
    def TriggerTimeMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # FaceAnimation
    def AnimName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def FaceAnimationStart(builder): builder.StartObject(2)
def FaceAnimationAddTriggerTimeMs(builder, triggerTimeMs): builder.PrependUint32Slot(0, triggerTimeMs, 0)
def FaceAnimationAddAnimName(builder, animName): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(animName), 0)
def FaceAnimationEnd(builder): return builder.EndObject()




from . import protocol_encoder


__all__ = [
    "Object",
]


class Object(object):

    def __init__(self, factory_id: int, object_type: protocol_encoder.ObjectType) -> None:
        self.factory_id = factory_id
        self.object_type = object_type



#!/usr/bin/env python

from typing import Optional
import sys
import time

import dpkt

import pycozmo


class ReplayApp(object):

    def __init__(self, log_messages: Optional[list] = None, replay_messages: Optional[list] = None):
        self.log_messages = log_messages
        self.frame_count = 0
        self.pkts = []
        self.first_ts = None
        self.packet_id_filter = pycozmo.filter.Filter()
        if replay_messages:
            for i in replay_messages:
                self.packet_id_filter.deny_ids(pycozmo.protocol_encoder.PACKETS_BY_GROUP[i])

    def load_engine_pkts(self, fspec):
        self.pkts = []

        self.frame_count = 0
        with open(fspec, "rb") as f:
            self.first_ts = None
            for ts, frame in dpkt.pcap.Reader(f):
                if self.first_ts is None:
                    self.first_ts = ts
                eth = dpkt.ethernet.Ethernet(frame)
                if eth.type != dpkt.ethernet.ETH_TYPE_IP:
                    # Skip non-IP frames
                    continue
                ip = eth.data
                if ip.p != dpkt.ip.IP_PROTO_UDP:
                    # Skip non-UDP frames
                    continue
                udp = ip.data
                if udp.data[:7] != pycozmo.protocol_declaration.FRAME_ID:
                    # Skip non-Cozmo frames
                    continue
                frame = pycozmo.Frame.from_bytes(udp.data)
                if frame.type not in (pycozmo.protocol_declaration.FrameType.ENGINE,
                                      pycozmo.protocol_declaration.FrameType.ENGINE_ACT):
                    # Skip non-engine frames
                    continue
                for pkt in frame.pkts:
                    if pkt.type not in [pycozmo.protocol_declaration.PacketType.COMMAND,
                                        pycozmo.protocol_declaration.PacketType.KEYFRAME]:
                        # Skip non-command packets
                        continue
                    self.pkts.append((ts, pkt))
                self.frame_count += 1

        print("Loaded {} engine packets from {} frames.".format(
            len(self.pkts), self.frame_count))

    def replay(self, fspec):
        self.load_engine_pkts(fspec)

        pycozmo.setup_basic_logging(log_level="DEBUG", protocol_log_level="DEBUG")
        cli = pycozmo.Client(protocol_log_messages=self.log_messages)
        cli.start()
        cli.connect()
        cli.wait_for_robot()

        try:
            for i, v in enumerate(self.pkts):
                # if i < 1113:
                #     continue
                ts, pkt = v
                if self.packet_id_filter.filter(pkt.id):
                    continue
                input()
                print("{}, time={:.06f}".format(i, ts - self.first_ts))
                cli.conn.send(pkt)
        except KeyboardInterrupt:
            pass

        cli.disconnect()
        time.sleep(1)


def main():
    fspec = sys.argv[1]
    log_messages = []   # "objects", "audio", "state"]
    replay_messages = []    # "lights", "objects"]

    app = ReplayApp(log_messages=log_messages, replay_messages=replay_messages)
    app.replay(fspec)


if __name__ == '__main__':
    main()



"""

Experimental code for reading Cozmo animations in .bin format.

Cozmo animations are stored in files/cozmo/cozmo_resources/assets/animations inside the Cozmo mobile application.

Animation data structures are declared in FlatBuffers format in files/cozmo/cozmo_resources/config/cozmo_anim.fbs .

"""

from typing import Optional
from collections import defaultdict
import math
import time

from PIL import Image
import numpy as np

from . import protocol_encoder
from . import lights
from . import procedural_face
from . import image_encoder
from . import anim_encoder
from . import robot


__all__ = [
    "PreprocessedClip",
]


class PreprocessedClip(object):

    def __init__(self, keyframes: Optional[defaultdict] = None):
        self.keyframes = keyframes or defaultdict(list)

    @classmethod
    def keyframe_to_im(cls, keyframe) -> Image:
        face = procedural_face.ProceduralFace(
            center_x=keyframe.center_x, center_y=keyframe.center_y,
            scale_x=keyframe.scale_x, scale_y=keyframe.scale_y,
            angle=keyframe.angle,
            left_eye=keyframe.left_eye, right_eye=keyframe.right_eye)
        im = face.render()
        # The Cozmo protocol expects a 128x32 image, so take only the even lines.
        np_im = np.array(im)
        np_im2 = np_im[::2]
        im = Image.fromarray(np_im2)
        return im

    @classmethod
    def from_anim_clip(cls, clip: anim_encoder.AnimClip):
        keyframes = defaultdict(list)
        for keyframe in clip.keyframes:
            if isinstance(keyframe, anim_encoder.AnimHeadAngle):
                # FIXME: Why can duration be larger than 255?
                pkt = protocol_encoder.AnimHead(duration_ms=min(keyframe.duration_ms, 255),
                                                variability_deg=keyframe.variability_deg,
                                                angle_deg=keyframe.angle_deg)
                keyframes[keyframe.trigger_time_ms].append(pkt)
            elif isinstance(keyframe, anim_encoder.AnimLiftHeight):
                # FIXME: Why can duration be larger than 255?
                pkt = protocol_encoder.AnimLift(duration_ms=min(keyframe.duration_ms, 255),
                                                variability_mm=keyframe.variability_mm,
                                                height_mm=keyframe.height_mm)
                keyframes[keyframe.trigger_time_ms].append(pkt)
            elif isinstance(keyframe, anim_encoder.AnimRecordHeading):
                pkt = protocol_encoder.RecordHeading()
                keyframes[keyframe.trigger_time_ms].append(pkt)
            elif isinstance(keyframe, anim_encoder.AnimTurnToRecordedHeading):
                pkt = protocol_encoder.TurnToRecordedHeading()
                keyframes[keyframe.trigger_time_ms].append(pkt)
            elif isinstance(keyframe, anim_encoder.AnimBodyMotion):
                if keyframe.radius_mm == "STRAIGHT":
                    pkt = protocol_encoder.AnimBody(speed=keyframe.speed, unknown=32767)
                elif keyframe.radius_mm == "TURN_IN_PLACE":
                    pkt = protocol_encoder.TurnInPlaceAtSpeed(wheel_speed_mmps=keyframe.speed,
                                                              direction=math.copysign(1.0, keyframe.speed))
                else:
                    assert isinstance(keyframe.radius_mm, float)
                    vl = keyframe.speed * (keyframe.radius_mm - robot.TRACK_WIDTH.mm / 2.0)
                    vr = keyframe.speed * (keyframe.radius_mm + robot.TRACK_WIDTH.mm / 2.0)
                    pkt = protocol_encoder.DriveWheels(lwheel_speed_mmps=vl, rwheel_speed_mmps=vr)
                keyframes[keyframe.trigger_time_ms].append(pkt)
                pkt = protocol_encoder.DriveWheels()
                keyframes[keyframe.trigger_time_ms + keyframe.duration_ms].append(pkt)
            elif isinstance(keyframe, anim_encoder.AnimBackpackLights):
                left = lights.Color(rgb=(keyframe.left.red, keyframe.left.green, keyframe.left.blue))
                front = lights.Color(rgb=(keyframe.front.red, keyframe.front.green, keyframe.front.blue))
                middle = lights.Color(rgb=(keyframe.middle.red, keyframe.middle.green, keyframe.middle.blue))
                back = lights.Color(rgb=(keyframe.back.red, keyframe.back.green, keyframe.back.blue))
                right = lights.Color(rgb=(keyframe.right.red, keyframe.right.green, keyframe.right.blue))
                pkt = protocol_encoder.AnimBackpackLights(colors=(left.to_int16(),
                                                                  front.to_int16(), middle.to_int16(), back.to_int16(),
                                                                  right.to_int16()))
                keyframes[keyframe.trigger_time_ms].append(pkt)
                off_light = lights.off.to_int16()
                pkt = protocol_encoder.AnimBackpackLights(colors=(off_light,
                                                                  off_light, off_light, off_light,
                                                                  off_light))
                keyframes[keyframe.trigger_time_ms + keyframe.duration_ms].append(pkt)
            elif isinstance(keyframe, anim_encoder.AnimFaceAnimation):
                # TODO
                pass
            elif isinstance(keyframe, anim_encoder.AnimProceduralFace):
                im = cls.keyframe_to_im(keyframe)
                encoder = image_encoder.ImageEncoder(im)
                buf = bytes(encoder.encode())
                pkt = protocol_encoder.DisplayImage(image=buf)
                keyframes[keyframe.trigger_time_ms].append(pkt)
            elif isinstance(keyframe, anim_encoder.AnimRobotAudio):
                # TODO
                pass
            elif isinstance(keyframe, anim_encoder.AnimEvent):
                # TODO
                pass
            else:
                raise RuntimeError("Unexpected keyframe type '{}'".format(type(keyframe)))
        ppclip = cls(keyframes=keyframes)
        return ppclip

    def play(self, cli):
        cli.conn.send(protocol_encoder.StartAnimation(anim_id=1))

        frames = list(sorted(self.keyframes.keys()))
        num_frames = len(frames)
        for i in range(num_frames):
            keyframe = self.keyframes[frames[i]]
            for action in keyframe:
                if isinstance(action, protocol_encoder.Packet):
                    cli.conn.send(action)

            # Play keyframe
            cli.conn.send(protocol_encoder.NextFrame())

            if i < num_frames - 1:
                delay_ms = (frames[i + 1] - frames[i]) / 1000.0
                time.sleep(delay_ms)

        cli.conn.send(protocol_encoder.EndAnimation())



"""

Helper routeines for working with colors and lights.

"""

from typing import Optional, Tuple

from . import protocol_encoder


__all__ = [
    'Color',

    'green',
    'red',
    'blue',
    'white',
    'off',

    'green_light',
    'red_light',
    'blue_light',
    'white_light',
    'off_light',
]


LED_ENC_BLUE = 0x001f
LED_ENC_GREEN = 0x03e0
LED_ENC_RED = 0x7c00
LED_ENC_IR = 0x8000

LED_ENC_BLUE_SHIFT = 0
LED_ENC_GREEN_SHIFT = 5
LED_ENC_RED_SHIFT = 10
LED_ENC_IR_SHIFT = 15


class Color:
    """
    A Color to be used with a Light.

    Either int_color or rgb may be used to specify the actual color.
    Any alpha components (from int_color) are ignored - all colors are fully opaque.

    Args:
        int_color (int): A 32 bit value holding the binary RGBA value.
        rgb (tuple): A tuple holding the integer values from 0-255 for (red, green, blue)
        name (str): A name to assign to this color
    """

    def __init__(self,
                 int_color: Optional[int] = None,
                 rgb: Optional[Tuple[int, int, int]] = None,
                 name: str = None) -> None:
        self.name = name
        if int_color is not None:
            self._int_color = int(int_color) | 0xff
        elif rgb is not None:
            self._int_color = (int(rgb[0]) << 24) | (int(rgb[1]) << 16) | (int(rgb[2]) << 8) | 0xff
        else:
            self._int_color = 0

    @property
    def int_color(self) -> int:
        return self._int_color

    def to_int16(self) -> int:
        r = ((self._int_color & 0xFF000000) >> 24) * 31 // 255
        g = ((self._int_color & 0x00FF0000) >> 16) * 31 // 255
        b = ((self._int_color & 0x0000FF00) >> 8) * 31 // 255
        value = (r << LED_ENC_RED_SHIFT) | (g << LED_ENC_GREEN_SHIFT) | (b << LED_ENC_BLUE_SHIFT)
        return value

    @classmethod
    def from_int16(cls, value: int):
        r = (value & LED_ENC_RED) >> LED_ENC_RED_SHIFT
        g = (value & LED_ENC_GREEN) >> LED_ENC_GREEN_SHIFT
        b = (value & LED_ENC_BLUE) >> LED_ENC_BLUE_SHIFT
        rgb = (
            r * 255 // 31,
            g * 255 // 31,
            b * 255 // 31
        )
        obj = cls(rgb=rgb)
        return obj

    def __repr__(self):
        return "Color(name={}, int_color=0x{:08x})".format(self.name, self._int_color)


green = Color(name="green", int_color=0x00ff00ff)
red = Color(name="red", int_color=0xff0000ff)
blue = Color(name="blue", int_color=0x0000ffff)
white = Color(name="white", int_color=0xffffffff)       # Does not work well with cubes?
off = Color(name="off")

green_light = protocol_encoder.LightState(on_color=green.to_int16(), off_color=green.to_int16())
red_light = protocol_encoder.LightState(on_color=red.to_int16(), off_color=red.to_int16())
blue_light = protocol_encoder.LightState(on_color=blue.to_int16(), off_color=blue.to_int16())
white_light = protocol_encoder.LightState(on_color=white.to_int16(), off_color=white.to_int16())
off_light = protocol_encoder.LightState(on_color=off.to_int16(), off_color=off.to_int16())



"""

Protocol AST.

"""

import enum
from abc import ABC
from typing import List, Optional, Union


class FrameType(enum.Enum):
    RESET = 1
    RESET_ACK = 2
    FIN = 3
    ENGINE_ACT = 4
    ENGINE = 7
    ROBOT = 9
    PING = 0x0b


class PacketType(enum.Enum):
    UNKNOWN = -1
    CONNECT = 2
    DISCONNECT = 3
    COMMAND = 4
    EVENT = 5
    KEYFRAME = 0x0a
    PING = 0x0b


class Argument(ABC):
    """ Base class for packet arguments. """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None):
        self.name = str(name) if name else None
        self.description = str(description) if description else None

    def type_hint(self) -> Optional[str]:
        return None


class FloatArgument(Argument):
    """ 32-bit floating point number. """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None, default: float = 0.0):
        super().__init__(name, description)
        self.default = float(default)

    def type_hint(self):
        return "float"


class DoubleArgument(Argument):
    """ 64-bit floating point number. """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None, default: float = 0.0):
        super().__init__(name, description)
        self.default = float(default)

    def type_hint(self):
        return "double"


class BoolArgument(Argument):
    """ 8-bit boolean. """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None, default: bool = False):
        super().__init__(name, description)
        self.default = bool(default)

    def type_hint(self):
        return "bool"


class UIntArgument(Argument, ABC):
    """ Base class for unsigned integers. """


class UInt8Argument(UIntArgument):
    """ 8-bit unsigned integer. """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None, default: int = 0):
        super().__init__(name, description)
        self.default = int(default)

    def type_hint(self):
        return "uint8"


class UInt16Argument(UIntArgument):
    """ 16-bit unsigned integer. """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None, default: int = 0):
        super().__init__(name, description)
        self.default = int(default)

    def type_hint(self):
        return "uint16"


class UInt32Argument(UIntArgument):
    """ 32-bit unsigned integer. """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None, default: int = 0):
        super().__init__(name, description)
        self.default = int(default)

    def type_hint(self):
        return "uint32"


class IntArgument(Argument, ABC):
    """ Base class for signed integers. """


class Int8Argument(IntArgument):
    """ 8-bit signed integer. """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None, default: int = 0):
        super().__init__(name, description)
        self.default = int(default)

    def type_hint(self):
        return "int8"


class Int16Argument(IntArgument):
    """ 16-bit signed integer. """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None, default: int = 0):
        super().__init__(name, description)
        self.default = int(default)

    def type_hint(self):
        return "int16"


class Int32Argument(IntArgument):
    """ 32-bit signed integer. """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None, default: int = 0):
        super().__init__(name, description)
        self.default = int(default)

    def type_hint(self):
        return "int32"


class EnumMember(object):
    """ Base class for enumeration members. """

    def __init__(self, name: str, value: int, description: Optional[str] = None):
        self.name = str(name)
        self.description = str(description) if description else None
        self.value = int(value)


class Enum(object):
    """ Base class for enumerations. """

    def __init__(self, name: str, description: Optional[str] = None,
                 members: Optional[List[EnumMember]] = None, base: int = 10):
        self.name = str(name)
        self.description = str(description) if description else None
        self.members = list(members) if members else []
        self.base = int(base)


class EnumArgument(Argument):
    """ Base class for enumeration arguments. """

    def __init__(self, name: str, enum_type: Enum, description: Optional[str] = None,
                 data_type: Union[IntArgument, UIntArgument] = Int8Argument(),
                 default=0):
        super().__init__(name, description)
        self.enum_type = enum_type
        self.data_type = data_type
        self.default = default

    def type_hint(self):
        return self.enum_type.name


class Struct(object):
    """ Base class for structures. """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None,
                 arguments: Optional[List[Argument]] = None):
        self.name = str(name) if name else None
        self.description = str(description) if description else None
        self.arguments = list(arguments) if arguments else []


class FArrayArgument(Argument):
    """ Fixed-length array. """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None,
                 data_type: Union[Argument, Struct] = UInt8Argument(), length: int = 0, default=()):
        super().__init__(name, description)
        self.data_type = data_type
        self.length = length
        self.default = tuple(default)

    def type_hint(self):
        if isinstance(self.data_type, Struct):
            type_name = self.data_type.name
        else:
            type_name = self.data_type.type_hint()
        return "{}[{}]".format(type_name, self.length)


class VArrayArgument(Argument):
    """ Variable-length array. """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None,
                 data_type: Argument = UInt8Argument(), length_type: Argument = UInt16Argument(), default=()):
        super().__init__(name, description)
        self.data_type = data_type
        self.length_type = length_type
        self.default = tuple(default)

    def type_hint(self):
        data_type_name = self.data_type.type_hint()
        length_type_name = self.length_type.type_hint()
        return "{}[{}]".format(data_type_name, length_type_name)


class StringArgument(Argument):
    """ String. """

    def __init__(self, name: Optional[str] = None, description: Optional[str] = None,
                 length_type: Argument = UInt16Argument(), default=""):
        super().__init__(name, description)
        self.length_type = length_type
        self.default = str(default)

    def type_hint(self):
        return "str"


class Packet(Struct, ABC):
    """ Base class for packets. """

    def __init__(self, packet_type: PacketType, name: str, packet_id: Optional[int] = None,
                 group: Optional[str] = None, description: Optional[str] = None,
                 arguments: Optional[List[Argument]] = None):
        super().__init__(name, description, arguments)
        self.type = PacketType(packet_type)
        self.id = packet_id
        self.group = group


class Connect(Packet):
    """ Connection acknowledgement packet. """

    def __init__(self):
        super().__init__(PacketType.CONNECT, "Connect")


class Disconnect(Packet):
    """ Disconnect packet. """

    def __init__(self):
        super().__init__(PacketType.DISCONNECT, "Disconnect")


class Command(Packet):
    """ Command packet. """

    def __init__(self, packet_id: int, name: str, group: Optional[str] = None,  description: Optional[str] = None,
                 arguments: Optional[List[Argument]] = None):
        super().__init__(PacketType.COMMAND, name, packet_id=packet_id, group=group, description=description,
                         arguments=arguments)


class Event(Packet):
    """ Event packet. """

    def __init__(self, packet_id: int, name: str, group: Optional[str] = None, description: Optional[str] = None,
                 arguments: Optional[List[Argument]] = None):
        super().__init__(PacketType.EVENT, name, packet_id=packet_id, group=group, description=description,
                         arguments=arguments)


class Ping(Packet):
    """ Ping packet. """

    def __init__(self):
        super().__init__(PacketType.PING, "Ping", arguments=[
            DoubleArgument("time_sent_ms"),
            UInt32Argument("counter"),
            UInt32Argument("last"),
            UInt8Argument("unknown"),
        ])


class Keyframe(Packet):

    def __init__(self):
        super().__init__(PacketType.KEYFRAME, "Keyframe")


class Protocol(object):
    """ Protocol declaration. """

    def __init__(self, enums: List[Enum], structs: List[Struct], packets: List[Packet]):
        self.enums = list(enums)
        self.structs = list(structs)
        self.packets = list(packets)




import sys

from .logging import *
from .run import *

from .frame import Frame
from .conn import ROBOT_ADDR
from .client import Client
from .robot import *
from .event import *

from . import exception
from . import util
from . import protocol_base
from . import protocol_declaration
from . import protocol_generator
from . import protocol_encoder
from . import protocol_utils
from . import lights
from . import camera
from . import object
from . import filter
from . import anim
from . import anim_encoder
from . import image_encoder
from . import procedural_face


__version__ = "0.6.0"

__all__ = [
    "logger",
    "logger_protocol",

    "Frame",
    "ROBOT_ADDR",
    "Client",

    "setup_basic_logging",
    "run_program",
]

if sys.version_info < (3, 5, 4):
    sys.exit("ERROR: PyCozmo requires Python 3.5.4 or newer.")
del sys
